#include <new>
#include <stdio.h>
#include <stdarg.h>
#include <squirrel.h>
#include <sqthread.h>
#include <sqcont.h>

#include <AS3.h>
#include "sqwrapper.h"
#include "sqas3obj.h"
#include "scfunc.h"

// ----------------------------------------------------------------

/**
 * squirrel用 print実装
 */
static void PrintFunc(HSQUIRRELVM v, const SQChar* format, ...)
{
	va_list args;
	va_start(args, format);
	vfprintf(stderr, format, args);
	va_end(args);
}

// ----------------------------------------------------------------------------
// 追加機能
// ----------------------------------------------------------------------------

// 引数エラー
SQRESULT ERROR_INVALIDPARAM(HSQUIRRELVM v) {
	return sq_throwerror(v, _SC("invalid param"));
}

/**
 * ステージを取得
 */
static SQRESULT getStage(HSQUIRRELVM vm)
{
	AS3Variant stage(AS3_Stage());
	return sq_pushvariant(vm, stage);
}

/**
 * トレースログ表示
 * @param ... 表示内容(文字列に変換されて表示されます)
 */
static SQRESULT trace(HSQUIRRELVM vm)
{
	SQInteger nargs = sq_gettop(vm);
	for (int i=2;i<=nargs;i++) {
		sq_tostring(vm, i);
		AS3Variant m = createVariant(vm, -1);
		m.trace();
		sq_pop(vm, 1);
	}
	return SQ_OK;
}

/**
 * Flash グローバルファンクション呼び出し (squirrel用)
 * @param namaspace
 * @param funcname
 * @param args 引数指定
 */
static SQRESULT callFlashFunction(HSQUIRRELVM vm)
{
	SQInteger top = sq_gettop(vm);
	if (top >= 2) {
		const char *funcName = NULL;
		if (SQ_SUCCEEDED(sq_getstring(vm, 2, &funcName))) {
			AS3Variant func = AS3Variant::createFunction(funcName);
			AS3Variant arg  = createArray(vm, 3, top);
			AS3Variant ret  = func.call(&arg);
			return sq_pushvariant(vm, ret);
		}
	}
	return ERROR_INVALIDPARAM(vm);
}

/**
 * Flash ファンクション生成
 */
static SQRESULT createFlashFunction(HSQUIRRELVM vm)
{
	if (sq_gettop(vm) >= 2) {
		AS3Variant func = createSCFunction(vm,2);
		return sq_pushvariant(vm, func);
	}
	return ERROR_INVALIDPARAM(vm);
}

/**
 * squirrel 初期化処理
 */
void sqas_init()
{
	HSQUIRRELVM vm = sqobject::init();
	sq_setprintfunc(vm, PrintFunc);

#ifdef USECONTINUOUS
	sqobject::registerContinuous();
#endif
	sqobject::Object::registerClass();
	sqobject::Thread::registerClass();
	sqobject::Thread::registerGlobal();

	AS3Object::init(vm);

	// グローバルメソッド・ファンクション登録
	
	sq_pushroottable(vm);

	// トレース用
	sq_pushstring(vm, _SC("trace"), -1);
	sq_newclosure(vm, trace, 0);
	sq_createslot(vm, -3); 

	// Flashグローバルファンクション呼び出し
	sq_pushstring(vm, _SC("callFlashFunction"), -1);
	sq_newclosure(vm, callFlashFunction, 0);
	sq_createslot(vm, -3); 

	// Flashファンクション生成
	sq_pushstring(vm, _SC("createFlashFunction"), -1);
	sq_newclosure(vm, createFlashFunction, 0);
	sq_createslot(vm, -3); 
	
	// ステージ取得用
	sq_pushstring(vm, _SC("getStage"), -1);
	sq_newclosure(vm, getStage, 0);
	sq_createslot(vm, -3); 

	sq_pop(vm, 1);

	scfunc_init();
	
	// ---------------------------
	// ActionScript 側への登録処理
	
	// 初期スクリプト起動
	sqobject::Thread::init();
}

/**
 * スレッド駆動
 * @param diff 経過時間
 * @return 駆動スレッド数
 */
int sqas_drive(int diff)
{
	int ret;
	static unsigned int scfuncTick = 0;
	sqobject::Thread::update(diff);
#ifdef USECONTINUOUS
	sqobject::beforeContinuous();
#endif
	ret = sqobject::Thread::main();
#ifdef USECONTINUOUS
	sqobject::afterContinuous();
#endif
	scfuncTick += diff;
	if (scfuncTick > 10000) {
		// 10秒に一回チェック
		scfunc_update();
		scfuncTick = 0;
	}
	return ret;
}

/**
 * squirrel 終了処理
 */
void sqas_done()
{
	scfunc_done();
#ifdef USECONTINUOUS
	sqobject::doneContinuous();
#endif
	sqobject::Thread::done();
	AS3Object::done();
	sqobject::done();
}

